<?php

namespace Drupal\taxonomy_term_replace\Form;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Defines A Confirmation Form Before Replacing Taxonomy Term.
 */
class TaxonomyTermConfirmationForm extends ConfirmFormBase {

  /**
   * The Request Stack Service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The EntityTypeManager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Database service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The Entity Field Manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Constructs a new TaxonomyTermConfirmationForm object.
   *
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   */
  public function __construct(
    RequestStack $request_stack,
    EntityTypeManagerInterface $entity_type_manager,
    Connection $connection,
    EntityFieldManagerInterface $entity_field_manager,
  ) {
    $this->requestStack = $request_stack;
    $this->entityTypeManager = $entity_type_manager;
    $this->connection = $connection;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new self(
      $container->get('request_stack'),
      $container->get('entity_type.manager'),
      $container->get('database'),
      $container->get('entity_field.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'taxonomy_term_confirmation';
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl() {
    return new Url('taxonomy_term_replace.dashboard');
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('This operation will put a significant load on the database. To ensure smooth processing, please make sure no other tasks are running while the terms are being replaced.<br> Would you like to continue?');
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return $this->t('<br> This Action Cannot Be Undone!');
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText() {
    return $this->t('Replace');
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelText() {
    return $this->t('Cancel');
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $request = $this->requestStack->getMainRequest();
    $illegal_confirmation = TRUE;
    $required_query_params = [
      'target_term_id',
      'replacement_term_id',
      'target_term',
      'replacement_term',
      'total_nodes',
    ];
    foreach ($required_query_params as $required_param) {
      if (empty($request->query->get($required_param))) {
        return new RedirectResponse(Url::fromRoute('taxonomy_term_replace.dashboard', [
          'illegal_confirmation' => $illegal_confirmation,
        ])->toString());
      }
    }
    $target_term_id = base64_decode($request->query->get('target_term_id'));
    $replacement_term_id = base64_decode($request->query->get('replacement_term_id'));
    $total_nodes = base64_decode($request->query->get('total_nodes'));
    $target_term = base64_decode($request->query->get('target_term'));
    $replacement_term = base64_decode($request->query->get('replacement_term'));
    $form['confirmation'] = [
      '#markup' => $this->t('You will be replacing term @target_term (Term ID  - @target_term_id) with @replacement_term (Term ID - @replacement_term_id) for a total of @total_nodes nodes associated to @target_term.', [
        '@target_term_id' => $target_term_id,
        '@replacement_term_id' => $replacement_term_id,
        '@total_nodes' => $total_nodes,
        '@target_term' => $target_term,
        '@replacement_term' => $replacement_term,
      ]),
    ];
    $form['#attributes']['class'][] = 'confirmation';
    // Add a description to provide additional context about the form page.
    $form['description'] = [
      '#markup' => $this->getDescription(),
    ];
    // Add a hidden form element to identify the form uniquely.
    $form[$this->getFormName()] = [
      '#type' => 'hidden',
      '#value' => 1,
    ];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->getConfirmText(),
      '#button_type' => 'primary',
    ];
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $request = $this->requestStack->getMainRequest();
    // Set Flag variable TRUE by default.
    $decoded_params_exist = TRUE;
    $query_params = [
      'target_term_id',
      'target_term',
      'replacement_term_id',
      'replacement_term',
      'unpublished',
      'selected_node_ids',
      'target_vocabulary',
    ];
    // Check if all required query parameters are present and set Flag variable
    // FALSE if even a single value is absent.
    foreach ($query_params as $param) {
      if (!isset($request->query->all()[$param])) {
        $decoded_params_exist = FALSE;
        break;
      }
    }
    if ($decoded_params_exist) {
      $taxonomy_storage = $this->entityTypeManager->getStorage('taxonomy_term');
      $node_bundle_storage = $this->entityTypeManager->getStorage('node_type');
      $node_storage = $this->entityTypeManager->getStorage('node');
      $target_term_id = base64_decode($request->query->get('target_term_id'));
      $replacement_term_id = base64_decode($request->query->get('replacement_term_id'));
      $target_term = base64_decode($request->query->get('target_term'));
      $replacement_term = base64_decode($request->query->get('replacement_term'));
      $unpublished = $request->query->get('unpublished');
      $json_decoded_selected_node_ids = base64_decode($request->query->get('selected_node_ids'));
      $selected_node_ids = Json::decode($json_decoded_selected_node_ids);
      $taxonomy_term = $taxonomy_storage->load($target_term_id);
      $vocabulary = $taxonomy_term->bundle();
      $target_vocabulary = base64_decode($request->query->get('target_vocabulary'));
      if (!$unpublished) {
        $query = $this->connection->select('taxonomy_index', 'ti');
        $query->fields('ti', ['nid']);
        $query->condition('ti.tid', $target_term_id, 'IN');
        $query->distinct(TRUE);
        $query->leftJoin('node_field_data', 'n', 'n.nid = ti.nid');
        $result = $query->execute();
        $nodes_details = $result->fetchCol();
      }
      else {
        $field_name = '';
        // Get all bundle/CT names.
        $all_bundles = $node_bundle_storage->loadMultiple();
        // Store only the machine name of the bundle(s)/Ct(s).
        $all_bundles = array_keys($all_bundles);
        foreach ($all_bundles as $bundle) {
          // Load all fields of the bundle/CT.
          $fields = $this->entityFieldManager->getFieldDefinitions('node', $bundle);
          foreach ($fields as $field) {
            // If field is a taxonomy term reference field.
            if (isset($field->getSettings()['target_type']) && $field->getSettings()['target_type'] == 'taxonomy_term') {
              $target_bundle = $field->getSettings()['handler_settings']['target_bundles'];
              // If target bundle of field is the same as the entered
              // term's corresponding vocabulary.
              if (array_key_first($target_bundle) == $target_vocabulary) {
                // Get field's name.
                $field_name = $field->getName();
              }
            }
          }
        }
        // If field exists.
        if ($field_name) {
          // Load nodes.
          $nodes_details = $node_storage->loadByProperties([$field_name => $target_term_id]);
          $nodes_details = array_keys($nodes_details);
        }
      }
      if ($nodes_details && (!empty($selected_node_ids))) {
        $nodes_details = array_intersect($nodes_details, $selected_node_ids);
        $nodes = $node_storage->loadMultiple($nodes_details);
        $loaded_vocab = $this->entityTypeManager->getStorage('taxonomy_vocabulary')->load($vocabulary);
        $vocabulary_label = !empty($loaded_vocab) ? $loaded_vocab->label() : $vocabulary;
        if ($nodes) {
          $batch = [
            'title' => $this->t('Replacing term - @target_term (Term ID -@target_term_id) in associated nodes with term - @replacement_term (Term ID - @replacement_term_id) belonging to @vocabulary vocabulary',
            [
              '@target_term_id' => $target_term_id,
              '@replacement_term_id' => $replacement_term_id,
              '@vocabulary' => $vocabulary_label,
              '@target_term' => $target_term,
              '@replacement_term' => $replacement_term,
            ],
            ),
            'init_message' => $this->t('Replacing'),
            'progress_message' => $this->t('Replacing terms in @current of @total Nodes.'),
            'error_message' => $this->t('An error occurred during replacement'),
            'finished' => '\Drupal\taxonomy_term_replace\Form\TaxonomyTermReplaceForm::taxonomyTermReplacementFinished',
          ];
          $batch_processing_chunk = array_chunk($nodes, 10);
          foreach ($batch_processing_chunk as $batch_processing_chunk_value) {
            foreach ($batch_processing_chunk_value as $individual_node) {
              // Get Field Name.
              $main_field = '';
              // Load All Fields Of The Node.
              $all_fields = $this->entityFieldManager->getFieldDefinitions('node', $individual_node->bundle());
              foreach ($all_fields as $field) {
                // If Field Is A Taxonomy Term Reference Field.
                $field_settings = $field->getSettings();
                if (isset($field_settings) && isset($field_settings['target_type']) && $field_settings['target_type'] == 'taxonomy_term') {
                  $target_bundle = $field->getSettings()['handler_settings']['target_bundles'];
                  // If Target Bundle Of Field Is The Same As The Entered Term's
                  // Corresponding Vocabulary.
                  if (array_key_first($target_bundle) == $vocabulary) {
                    // Get Field's Name.
                    $field_name = $field->getName();
                    // Get Field's Value.
                    $field_value = $individual_node->get($field_name)->getValue();
                    // If Field Value Is Not Empty.
                    if ($field_value) {
                      // Get The First & Only Value Of Field.
                      $field_value = reset($field_value[0]);
                      // If Field Value Is Equal To Entered Term ID, i.e.
                      // Target Field.
                      if ($field_value == $target_term_id) {
                        $main_field = $field_name;
                      }
                    }
                  }
                }
              }
              $batch['operations'][] = ['\Drupal\taxonomy_term_replace\Form\TaxonomyTermReplaceForm::taxonomyTermReplacement',
                [
                  $individual_node,
                  $vocabulary_label,
                  $target_term_id,
                  $replacement_term_id,
                  $main_field,
                  $target_term,
                  $replacement_term,
                ],
              ];
            }
          }
          if (!empty($batch['operations'])) {
            batch_set($batch);
          }
        }
      }
    }
    $form_state->setRedirect('taxonomy_term_replace.dashboard');
  }

}
