<?php

namespace Drupal\taxonomy_term_replace\Form;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Url;
use Drupal\node\NodeInterface;
use Drupal\taxonomy\TermInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Form for Taxonomy Term Replace Dashboard for associated nodes for a term.
 */
class TaxonomyTermReplaceForm extends FormBase {

  /**
   * The EntityTypeManager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Database service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The Request Stack Service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The Entity Field Manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Constructs a new TaxonomyTermReplaceForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    Connection $connection,
    MessengerInterface $messenger,
    RequestStack $request_stack,
    EntityFieldManagerInterface $entity_field_manager,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->connection = $connection;
    $this->messenger = $messenger;
    $this->requestStack = $request_stack;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new self(
      $container->get('entity_type.manager'),
      $container->get('database'),
      $container->get('messenger'),
      $container->get('request_stack'),
      $container->get('entity_field.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'taxonomy_term_replace';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Get request to check whether it is an illegal confirmation or not, i.e.,
    // User has moved to the confirmation page directly.
    $request = $this->requestStack->getMainRequest();
    // Check for illegal confirmation and redirect if necessary.
    if ($request->query->has('illegal_confirmation')) {
      $this->messenger->addError($this->t('Add "Target" and "Replacement" terms before proceeding to replacement procedure!'));
      return new RedirectResponse(Url::fromRoute('<current>')->toString());
    }
    // Load all available vocabularies.
    $vocabularies = $this->entityTypeManager->getStorage('taxonomy_vocabulary')->loadMultiple();
    $options = [];
    foreach ($vocabularies as $vocabulary) {
      // Using vocabulary ID as the key, and the label as the value, i.e.
      // Key:Value => Vocabulary_ID:Vocabulary_Label, create Vocabularies array.
      $options[$vocabulary->id()] = $vocabulary->label();
    }
    $form['vocabulary'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Select vocabulary here'),
      '#open' => TRUE,
    ];
    // Vocabulary field.
    $form['vocabulary']['vocabulary_main'] = [
      '#type' => 'select',
      '#title' => $this->t('Select vocabulary'),
      '#options' => $options,
      '#required' => TRUE,
      '#ajax' => [
        'callback' => [$this, 'loadTermsCallback'],
        'wrapper' => 'taxonomy-terms-wrapper',
        'progress' => [
          'type' => 'throbber',
          'message' => $this->t('Loading terms...'),
        ],
      ],
    ];
    // Render taxonomy terms wrapper field - will be visible when any
    // vocabularies are selected.
    $form['taxonomy_terms'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Select taxonomy terms here'),
      '#open' => TRUE,
      '#attributes' => [
        'id' => 'taxonomy-terms-wrapper',
      ],
      '#states' => [
        'visible' => [
          ':input[name="vocabulary_main"]' => ['!value' => ''],
        ],
      ],
    ];
    // Get selected vocabularies from field.
    $selected_vocabulary = $form_state->getValue('vocabulary_main');
    $taxonomy_data = $this->entityTypeManager->getStorage('taxonomy_term');
    if (!empty($selected_vocabulary)) {
      // Load entire selected vocabulary.
      $terms = $taxonomy_data->loadTree(
        $selected_vocabulary,
        0,
        NULL,
        TRUE,
      );
      if ($terms) {
        $term_options = [];
        // Using term ID as the key, and the name as the value, i.e.
        // Key:Value => Term_ID:Term_Name, create taxonomy terms array.
        foreach ($terms as $term) {
          $term_options[$term->id()] = $term->getName();
        }
        // Render taxonomy terms selection fields.
        if (!empty($term_options)) {
          // Render target term field.
          $form['taxonomy_terms']['target_term'] = [
            '#type' => 'select',
            '#title' => $this->t('Select target term'),
            '#options' => $term_options,
            '#empty_option' => $this->t('- Select -'),
            '#ajax' => [
              'callback' => [$this, 'renderActionButtons'],
              'wrapper' => 'action-buttons-wrapper',
            ],
          ];
          // Render replacement term field.
          $form['taxonomy_terms']['replacement_term'] = [
            '#type' => 'select',
            '#title' => $this->t('Select replacement term'),
            '#options' => $term_options,
            '#empty_option' => $this->t('- Select -'),
          ];
          $form_state->setRebuild(TRUE);
        }
      }
      else {
        $form['taxonomy_terms']['no_terms_message'] = [
          '#markup' => $this->t('Please select a vocabulary with existing terms in it before proceeding.'),
        ];
        $form_state->setRebuild(TRUE);
      }
    }
    // Render action buttons when target term field has a selection.
    $form['action_buttons'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'action-buttons-wrapper'],
      '#states' => [
        'visible' => [
          ':input[name="target_term"]' => ['!value' => ''],
        ],
      ],
    ];
    // Get selected target term from field.
    $selected_target_term = $form_state->getValue('target_term');
    if ($selected_target_term) {
      if ($target_term = $taxonomy_data->load($selected_target_term)) {
        if ($target_term instanceof TermInterface) {
          // Get name of selected target term.
          $selected_term_name = $target_term->getName() ?? '';
        }
      }
      $form['action_buttons']['associated_nodes'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Search nodes for term - @selected_term_name', [
          '@selected_term_name' => $selected_term_name,
        ]),
      ];
      // Button for adding unpublished nodes in the process as well.
      $form['action_buttons']['associated_nodes']['unpublished_nodes'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Add unpublished nodes'),
        '#description' => $this->t('Adds unpublished nodes to both the search results and further replacement process.'),
      ];
      // Render search button for all associated nodes to the target term.
      $form['action_buttons']['associated_nodes']['actions']['search'] = [
        '#type' => 'submit',
        '#value' => $this->t('Search'),
        '#button_type' => 'secondary',
        '#submit' => [
          [$this, 'searchContentBasedOnTermId'],
        ],
      ];
      $form_state->setRebuild(TRUE);
    }
    // Set nodes data in array when 'Search' button is clicked.
    $nodes_data = [];
    $nodes_data = $form_state->get('nodes_data');
    // If nodes' data are present in array, render table to view all nodes.
    if ($nodes_data) {
      // Create Download Table button if nodes exist.
      $form['action_buttons']['associated_nodes']['actions']['csv_download'] = [
        '#type' => 'submit',
        '#value' => $this->t('Download table'),
        '#button_type' => 'secondary',
        '#submit' => [
          [$this, 'downloadCsv'],
        ],
      ];
      $form['action_buttons']['associated_nodes']['#title'] = $this->t('Check below results and download a CSV file corresponding to nodes belonging to searched term');
      $form['selected_nodes'] = [
        '#type' => 'fieldset',
        '#title' => $this->t('Nodes associated to searched "Target" term - @selected_term_name', [
          '@selected_term_name' => $selected_term_name,
        ]),
      ];
      $node_options = [];
      foreach ($nodes_data as $node) {
        if ($node instanceof NodeInterface) {
          $node_url = $node->toUrl()->setOption('absolute', TRUE);
          $node_options[$node->id()] = [
            'node_id' => $node->id(),
            'node_url' => Link::fromTextAndUrl($node->getTitle(), $node_url)->toString(),
            'node_type' => $node->type->entity->label(),
            'node_status' => $node->isPublished() ? 'Published' : 'Unpublished',
          ];
        }
      }
      $form['selected_nodes']['details'] = [
        '#type' => 'details',
        '#title' => $this->t('Listed below are a total of @total_nodes nodes associated with selected "Target" term. Please choose any/all rows from below table in order to proceed to replacement process.', [
          '@total_nodes' => count($nodes_data),
        ]),
        '#open' => TRUE,
      ];
      // Table renders - Node's ID, Node's Title / URL, Node's Content Type,
      // Node's Status.
      // Render data in 'tableselect' type field in order to have specific / all
      // node selection from user before proceeding to replacement process.
      $form['selected_nodes']['details']['existing_nodes'] = [
        '#type' => 'tableselect',
        '#header' => [
          'node_id' => $this->t('Node ID'),
          'node_url' => $this->t('Node URL'),
          'node_type' => $this->t('Content Type'),
          'node_status' => $this->t('Node Status'),
        ],
        '#options' => $node_options,
        '#empty' => $this->t('No nodes found.'),
      ];
      // Render process replacement button.
      $form['selected_nodes']['actions']['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Process replacement'),
        '#button_type' => 'primary',
        '#states' => [
          'visible' => [
            ':input[name="existing_nodes"]' => ['!value' => ''],
          ],
        ],
      ];
      // Render Reset Button if nodes exist.
      $form['action_buttons']['associated_nodes']['actions']['reset'] = [
        '#type' => 'submit',
        '#value' => $this->t('Reset'),
        '#button_type' => 'primary',
        '#submit' => ['::resetButton'],
        '#limit_validation_errors' => [],
        '#weight' => 10,
      ];
    }
    return $form;
  }

  /**
   * Creates 'Target' & 'Replacement' fields based on selected vocabulary.
   *
   * @param array $form
   *   Stores the data in the form fields.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Stores the object of FormStateInterface.
   *
   * @return array
   *   The form elements for taxonomy terms.
   */
  public function loadTermsCallback(array &$form, FormStateInterface $form_state) {
    return $form['taxonomy_terms'];
  }

  /**
   * Creates 'Unpublished Nodes' Checkbox & 'Search' Button.
   *
   * @param array $form
   *   Stores the data in the form fields.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Stores the object of FormStateInterface.
   *
   * @return array
   *   The form elements for action buttons.
   */
  public function renderActionButtons(array &$form, FormStateInterface $form_state) {
    return $form['action_buttons'];
  }

  /**
   * This method validates the 'Target' term & lists all the nodes linked to it.
   *
   * @param array $form
   *   Stores the data in the form fields.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Stores the object of FormStateInterface.
   *
   * @return array
   *   The form elements for selected nodes.
   */
  public function searchContentBasedOnTermId(array &$form, FormStateInterface $form_state) {
    $form['selected_nodes'] = [];
    $input_element = $form_state->cleanValues()->getValues();
    $input_search = $input_element['target_term'];
    $unpublished_nodes = $input_element['unpublished_nodes'];
    $taxonomy_data = $this->entityTypeManager->getStorage('taxonomy_term');
    $node_data = $this->entityTypeManager->getStorage('node');
    $node_bundle_data = $this->entityTypeManager->getStorage('node_type');
    if (isset($form['selected_nodes']['details']) && is_array($form['selected_nodes']['details'])) {
      unset($form['selected_nodes']['details']['existing_nodes']);
    }
    // Checking whether the input Term is empty or not.
    if (!empty($input_search)) {
      // Checking whether the input Term's ID is numeric or not.
      if (is_numeric($input_search)) {
        $taxonomy_term = $taxonomy_data->load($input_search);
        // Checking whether the input Term is a valid taxonomy term or not.
        if ($taxonomy_term) {
          $vocabulary = $taxonomy_term->bundle();
          // Checking whether the input Term belongs to a valid vocabulary
          // or not.
          if ($vocabulary) {
            $term_name = $taxonomy_term->label();
            // Checking whether the input Term is valid or not.
            if ($term_name) {
              // Initialize empty array for storing nodes.
              $nodes_details = [];
              // Only published nodes.
              if (!$unpublished_nodes) {
                $query = $this->connection->select('taxonomy_index', 'ti');
                $query->fields('ti', ['nid']);
                $query->condition('ti.tid', $input_search, 'IN');
                $query->distinct(TRUE);
                $query->leftJoin('node_field_data', 'n', 'n.nid = ti.nid');
                $result = $query->execute();
                $nodes_details = $result->fetchCol();
              }
              else {
                $field_name = '';
                // Get all bundle/CT names.
                $all_bundles = $node_bundle_data->loadMultiple();
                // Store only the machine name of the bundle(s)/Ct(s).
                $all_bundles = array_keys($all_bundles);
                foreach ($all_bundles as $bundle) {
                  // Load all fields of the bundle/CT.
                  $fields = $this->entityFieldManager->getFieldDefinitions('node', $bundle);
                  foreach ($fields as $field) {
                    // If field is a taxonomy term reference field.
                    if (isset($field->getSettings()['target_type']) && $field->getSettings()['target_type'] == 'taxonomy_term') {
                      $target_bundle = $field->getSettings()['handler_settings']['target_bundles'];
                      // If target bundle of field is the same as the entered
                      // term's corresponding vocabulary.
                      if (array_key_first($target_bundle) == $vocabulary) {
                        // Get field's name.
                        $field_name = $field->getName();
                      }
                    }
                  }
                }
                // If field exists.
                if ($field_name) {
                  // Load nodes.
                  $nodes_details = $node_data->loadByProperties([$field_name => $input_search]);
                  $nodes_details = array_keys($nodes_details);
                }
              }
              // Checking whether there are any existing nodes or not.
              if ($nodes_details) {
                $nodes = $node_data->loadMultiple($nodes_details);
                // Checking whether the input Term is associated with any
                // existing Nodes or not.
                if ($nodes) {
                  // Rebuild form with selected values, along with node table.
                  $form_state->set('nodes_data', $nodes);
                  $form_state->setRebuild(TRUE);
                  $form_state->disableRedirect(TRUE);
                }
                else {
                  $this->messenger->addError('Selected "Target" term has no known associated nodes to it. Please re-select a different term & try again.');
                }
              }
              else {
                $this->messenger->addError('Selected "Target" term has no known associated nodes to it. Please re-select a different term & try again.');
              }
            }
            else {
              $this->messenger->addError('Please select valid "Target" term.');
            }
          }
          else {
            $this->messenger->addError('"Target" term belongs to invalid / non-existent vocabulary.');
          }
        }
        else {
          $this->messenger->addError('Please select valid "Target" term.');
        }
      }
      else {
        $this->messenger->addError('Please select valid "Target" term.');
      }
    }
    else {
      $this->messenger->addError('"Target" term cannot be empty.');
    }
    return $form['selected_nodes'];
  }

  /**
   * Resets The Form Back To Original/Normal.
   *
   * @param array $form
   *   Stores The Form In Array Format.
   * @param Drupal\Core\Form\FormStateInterface $form_state
   *   Stores The Form State.
   */
  public function resetButton(array &$form, FormStateInterface $form_state) {
    $form_state->cleanValues();
    $return_url = Url::fromRoute('<current>', [], []);
    $form_state->setRedirectUrl($return_url);
  }

  /**
   * This Method replaces 'target' Term with selected 'replacement' term.
   *
   * @param object $individual_node
   *   Stores the individual nodes.
   * @param string $vocabulary
   *   Stores the Vocabulary of the terms.
   * @param int $target_term_id
   *   Stores the existing Term ID.
   * @param int $replacement_term_id
   *   Stores the replacement Term ID.
   * @param string $main_field
   *   Stores the Target Field Name.
   * @param string $target_term
   *   Stores the Target Term Name.
   * @param string $replacement_term
   *   Stores the Replacement Term Name.
   * @param array $context
   *   Stores The Batch Context.
   */
  public static function taxonomyTermReplacement(object $individual_node, string $vocabulary, int $target_term_id, int $replacement_term_id, string $main_field, string $target_term, string $replacement_term, array &$context) {
    // If main field is not empty.
    if ($main_field) {
      // If node has field & field is not empty.
      if ($individual_node->hasField($main_field)) {
        if (!empty($individual_node->get($main_field)->getValue())) {
          // Store existing terms from nodes.
          $transferred_value = $individual_node->get($main_field)->getValue();
          foreach ($transferred_value as $index => $transferred_replace) {
            if ($transferred_replace['target_id'] == $target_term_id) {
              $transferred_value[$index]['target_id'] = $replacement_term_id;
              break;
            }
          }
          // Set replacement term in place of existing target term for nodes.
          $individual_node->set($main_field, $transferred_value);
          // Save node once replacement is completed.
          $individual_node->save();
          // Store the total nodes' ID where replacement of term was successful
          // in the results array.
          $context['results']['count_nodes'][] = $individual_node->id();
        }
      }
    }
    // Store replacement process' data in $context.
    $context['results']['target_term'] = $target_term;
    $context['results']['target_term_id'] = $target_term_id;
    $context['results']['replacement_term'] = $replacement_term;
    $context['results']['replacement_term_id'] = $replacement_term_id;
    $context['results']['vocabulary'] = $vocabulary;
  }

  /**
   * This Method is called once term replacement process has finished.
   */
  public static function taxonomyTermReplacementFinished($success, $results, $operations) {
    if ($success) {
      $message = t('@count nodes processed. @target_term (Term ID - @target_term_id) term replaced with @replacement_term (Term ID - @replacement_term_id) belonging to @vocabulary vocabulary', [
        '@count' => count($results['count_nodes']),
        '@target_term' => $results['target_term'],
        '@target_term_id' => $results['target_term_id'],
        '@replacement_term' => $results['replacement_term'],
        '@replacement_term_id' => $results['replacement_term_id'],
        '@vocabulary' => $results['vocabulary'],
      ]);
      \Drupal::messenger()->addStatus($message);
    }
    else {
      $message = t('Finished with an error.');
      \Drupal::messenger()->addError($message);
    }
  }

  /**
   * Downloads Search Result In CSV Format Via Batch Processing.
   *
   * @param array $form
   *   Stores The Form In Array Format.
   * @param Drupal\Core\Form\FormStateInterface $form_state
   *   Stores The Form State.
   */
  public function downloadCsv(array &$form, FormStateInterface $form_state) {
    $nodes_data = [];
    // Get nodes' data.
    $nodes_data = $form_state->get('nodes_data');
    $target_term_id = $form_state->cleanValues()->getValues()['target_term'];
    $taxonomy_data = $this->entityTypeManager->getStorage('taxonomy_term');
    $target_taxonomy_term = $taxonomy_data->load($target_term_id)->label();
    // If nodes exist, create and download CSV via batch process.
    if (!empty($nodes_data)) {
      $batch = [
        'title' => $this->t('Generating CSV file for searched data! Term - @target_term', [
          '@target_term' => $target_taxonomy_term,
        ]),
        'init_message' => $this->t('Generating CSV'),
        'progress_message' => $this->t('Nodes processed - @current of @total nodes'),
        'error_message' => $this->t('An error occurred during processing nodes & generating CSV file!'),
        'finished' => '\Drupal\taxonomy_term_replace\Form\TaxonomyTermReplaceForm::generatingCsvCompleted',
      ];
      // Put nodes' data in chunks of 50.
      $batch_node = array_chunk($nodes_data, 50);
      foreach ($batch_node as $batch_node_chunk) {
        foreach ($batch_node_chunk as $batch_node) {
          $batch['operations'][] = ['\Drupal\taxonomy_term_replace\Form\TaxonomyTermReplaceForm::generateCsv',
          [
            $batch_node,
          ],
          ];
        }
      }
      if (!empty($batch['operations'])) {
        batch_set($batch);
      }
    }
  }

  /**
   * Gets Required Information From Nodes And Generates CSV File.
   *
   * @param object $node
   *   Stores The Individual Nodes.
   * @param array $context
   *   Stores The Batch Context.
   */
  public static function generateCsv(object $node, array $context) {
    $node_data = [];
    $node_url = $node->toUrl()->setOption('absolute', TRUE)->toString();
    $node_data[] = [
      $node->id(),
      $node_url,
      $node->bundle(),
      $node->isPublished() ? 'Published' : 'Unpublished',
    ];
    $context['results']['node_data'][] = $node_data;
  }

  /**
   * Method Called When CSV Generation Process Is Completed.
   */
  public static function generatingCsvCompleted($success, $results, $operations) {
    if ($success) {
      // Prepare CSV data.
      $filename = 'public://taxonomy-term-node-search.csv';
      $handle = fopen($filename, 'w');
      $header = [
        'Node ID',
        'Node URL',
        'Content Type',
        'Node Status',
      ];
      // Put data in CSV file.
      fputcsv($handle, $header);
      foreach ($results['node_data'] as $result) {
        foreach ($result as $value) {
          fputcsv($handle, array_values($value));
        }
      }
      fclose($handle);
      // Download CSV.
      if (file_exists($filename)) {
        $message = 'CSV file successfully downloaded.';
        $result = TRUE;
        // Set headers for the file download.
        header('Content-Description: File Transfer');
        header('Content-Type: application/octet-stream');
        header('Content-Disposition: attachment; filename="' . basename($filename) . '"');
        header('Expires: 0');
        header('Cache-Control: must-revalidate');
        header('Pragma: public');
        header('Content-Length: ' . filesize($filename));
        readfile($filename);
        exit;
      }
    }
    else {
      $message = t('Finished with an error.');
      $result = FALSE;
    }
    if ($result) {
      \Drupal::messenger()->addMessage($message);
    }
    else {
      \Drupal::messenger()->addError($message);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Get value of the rows selected after node search based on 'target' term.
    $tableselect_values = $form_state->getValue('existing_nodes');
    // Filter out only the selected rows.
    $selected_node_ids = [];
    if (isset($tableselect_values)) {
      foreach ($tableselect_values as $index => $selected) {
        if ($selected) {
          // Store selected node IDs in array.
          $selected_node_ids[] = $index;
        }
      }
    }
    $input_element = $form_state->cleanValues()->getValues();
    // Store entered 'target' term.
    if (isset($input_element['target_term'])) {
      $target_term_id = $input_element['target_term'];
    }
    // Store entered 'replacement' term.
    if (isset($input_element['replacement_term'])) {
      $replacement_term_id = $input_element['replacement_term'];
    }
    // Store 'Add Unpublished Nodes' checkbox input.
    if (isset($input_element['unpublished_nodes'])) {
      $unpublished_nodes = $input_element['unpublished_nodes'];
    }
    $taxonomy_data = $this->entityTypeManager->getStorage('taxonomy_term');
    $node_data = $this->entityTypeManager->getStorage('node');
    $node_bundle_storage = $this->entityTypeManager->getStorage('node_type');
    // Checking whether 'target' & 'replacement' terms are empty or not.
    if (!empty($target_term_id) && !empty($replacement_term_id)) {
      // Checking whether 'target' & 'replacement' terms' ID are numeric or not.
      if (is_numeric($target_term_id) && (is_numeric($replacement_term_id))) {
        $target_taxonomy_term = $taxonomy_data->load($target_term_id);
        $replacement_taxonomy_term = $taxonomy_data->load($replacement_term_id);
        // Checking whether 'target' & 'replacement' terms' are valid or not.
        if ((!empty($target_taxonomy_term)) && (!empty($replacement_taxonomy_term))) {
          $target_vocabulary = $target_taxonomy_term->bundle();
          $replacement_vocabulary = $replacement_taxonomy_term->bundle();
          // Checking whether 'target' & 'replacement' term's vocabulary exist
          // or not.
          if (!empty($target_vocabulary) && !empty($replacement_vocabulary)) {
            // Checking whether 'target' & 'replacement' term's vocabulary are
            // same or not.
            if ($target_vocabulary == $replacement_vocabulary) {
              // Checking whether 'target' & 'replacement' terms are different
              // or not.
              if ($target_term_id <> $replacement_term_id) {
                // Initialize empty array for storing nodes.
                $nodes_details = [];
                // Only published nodes.
                if (!$unpublished_nodes) {
                  $query = $this->connection->select('taxonomy_index', 'ti');
                  $query->fields('ti', ['nid']);
                  $query->condition('ti.tid', $target_term_id, 'IN');
                  $query->distinct(TRUE);
                  $query->leftJoin('node_field_data', 'n', 'n.nid = ti.nid');
                  $result = $query->execute();
                  $nodes_details = $result->fetchCol();
                }
                // All nodes, including unpublished Ones.
                else {
                  $field_name = '';
                  // Get all bundle/CT names.
                  $all_bundles = $node_bundle_storage->loadMultiple();
                  // Store only the machine name of the bundle(s)/Ct(s).
                  $all_bundles = array_keys($all_bundles);
                  foreach ($all_bundles as $bundle) {
                    // Load all fields of the bundle/CT.
                    $fields = $this->entityFieldManager->getFieldDefinitions('node', $bundle);
                    foreach ($fields as $field) {
                      // If field is a taxonomy term reference field.
                      if (isset($field->getSettings()['target_type']) && $field->getSettings()['target_type'] == 'taxonomy_term') {
                        $target_bundle = $field->getSettings()['handler_settings']['target_bundles'];
                        // If target bundle of field is the same as the entered
                        // term's corresponding vocabulary.
                        if (array_key_first($target_bundle) == $target_vocabulary) {
                          // Get field's name.
                          $field_name = $field->getName();
                        }
                      }
                    }
                  }
                  // If field exists.
                  if ($field_name) {
                    // Load nodes.
                    $nodes_details = $node_data->loadByProperties([$field_name => $target_term_id]);
                    $nodes_details = array_keys($nodes_details);
                  }
                }
                // Checking whether query returned any valid Node IDs or not.
                if ($nodes_details) {
                  // Checking whether there are selected rows or not from the
                  // search result.
                  if (!empty($selected_node_ids)) {
                    // Selecting only those nodes that have been selected from
                    // the search results.
                    $nodes_details = array_intersect($nodes_details, $selected_node_ids);
                    $nodes = $node_data->loadMultiple($nodes_details);
                    $json_encoded_selected_node_ids = Json::encode($selected_node_ids);
                    // Checking whether there are nodes associated with entered
                    // 'target' term or not.
                    if ($nodes) {
                      // Redirect user to confirmation page for the final
                      // replacement process.
                      $form_state->setRedirect('taxonomy_term_replace.confirmation', [
                        'target_term_id' => base64_encode($target_term_id),
                        'replacement_term_id' => base64_encode($replacement_term_id),
                        'total_nodes' => base64_encode(count($nodes)),
                        'target_term' => base64_encode($target_taxonomy_term->label()),
                        'replacement_term' => base64_encode($replacement_taxonomy_term->label()),
                        'unpublished' => $unpublished_nodes,
                        'selected_node_ids' => base64_encode($json_encoded_selected_node_ids),
                        'target_vocabulary' => base64_encode($target_vocabulary),
                      ]);
                    }
                    else {
                      $this->messenger->addError('Selected term has no known associated nodes to it. Please re-select term & try again.');
                    }
                  }
                  else {
                    $this->messenger->addError('Please select any/all rows from the search result before proceeding to term replacement.');
                  }
                }
                else {
                  $this->messenger->addError('Selected term has no known associated nodes to it. Please re-select term & try again.');
                }
              }
              else {
                $this->messenger->addError('"Target" & "Replacement" terms cannot be the same.');
              }
            }
            else {
              $this->messenger->addError('"Term(s)" must belong to the same vocabulary.');
            }
          }
          else {
            $this->messenger->addError('"Term(s)" belong to invalid / non-existent vocabulary.');
          }
        }
        else {
          $this->messenger->addError('Invalid term(s).');
        }
      }
      else {
        $this->messenger->addError('Invalid term(s).');
      }
    }
    else {
      $this->messenger->addError('Please select terms for both "Target" & "Replacement".');
    }
  }

}
